@@ -1,7 +1,7 @@
 /* -*- indent-tabs-mode: t; tab-width: 8; c-basic-offset: 8; -*- */
 
 /* Copyright (c) 2004 - 2006 Derek Foreman, Ben Jansens
-   Copyright (c) 2006 - 2012 Thomas Schmitt <scdbackup@gmx.net>
+   Copyright (c) 2006 - 2017 Thomas Schmitt <scdbackup@gmx.net>
    Provided under GPL version 2 or later.
 */
 
@@ -98,6 +98,14 @@ struct fifo_opts
 	int flag;
 };
 
+union w_list_data
+{
+	struct scan_opts scan;
+	struct erase_opts erase;
+	struct format_opts format;
+	struct write_opts write;
+	struct fifo_opts fifo;
+};
 
 struct w_list
 {
@@ -109,19 +117,48 @@ struct w_list
 
 	struct w_list *next;
 
-	union w_list_data
-	{
-		struct scan_opts scan;
-		struct erase_opts erase;
-		struct format_opts format;
-		struct write_opts write;
-		struct fifo_opts fifo;
-	} u;
+	union w_list_data u;
 };
 
 static struct w_list *workers = NULL;
 
 
+int burn_async_manage_lock(int mode)
+{
+	int ret;
+
+	static pthread_mutex_t access_lock;
+	static int mutex_initialized = 0;
+	static int mutex_locked = 0;
+
+	if (mode == BURN_ASYNC_LOCK_INIT) {
+		if (mutex_initialized)
+			return 2;
+		ret = pthread_mutex_init(&access_lock, NULL);
+		if (ret != 0)
+			return 0;
+		mutex_initialized = 1;
+		return 1;
+	}
+	if (!mutex_initialized)
+		return 0;
+	if (mode == BURN_ASYNC_LOCK_OBTAIN) {
+		ret = pthread_mutex_lock(&access_lock);
+		if (ret != 0)
+			return 0;
+		mutex_locked = 1;
+	} else if (mode == BURN_ASYNC_LOCK_RELEASE) {
+		if (!mutex_locked)
+			return 2;
+		ret = pthread_mutex_unlock(&access_lock);
+		if (ret != 0)
+			return 0;
+		mutex_locked = 0;
+	}
+	return 1;
+}
+
+
 static struct w_list *find_worker(struct burn_drive *d)
 {
 	struct w_list *a;
@@ -133,7 +170,7 @@ static struct w_list *find_worker(struct burn_drive *d)
 }
 
 static void add_worker(int w_type, struct burn_drive *d,
-			WorkerFunc f, void *data)
+			WorkerFunc f, union w_list_data *data)
 {
 	struct w_list *a;
 	struct w_list *tmp;
@@ -146,7 +183,10 @@ static void add_worker(int w_type, struct burn_drive *d,
 	a = calloc(1, sizeof(struct w_list));
 	a->w_type = w_type;
 	a->drive = d;
-	a->u = *(union w_list_data *)data;
+
+	a->u = *data;
+
+	burn_async_manage_lock(BURN_ASYNC_LOCK_INIT);
 
 	/* insert at front of the list */
 	a->next = workers;
@@ -157,6 +197,7 @@ static void add_worker(int w_type, struct burn_drive *d,
 		d->busy = BURN_DRIVE_SPAWNING;
 
 #ifdef Libburn_create_detached_threadS
+
 	/* ts A71019 :
 	   Trying to start the threads detached to get rid of the zombies
 	   which do neither react on pthread_join() nor on pthread_detach().
@@ -164,12 +205,12 @@ static void add_worker(int w_type, struct burn_drive *d,
 	pthread_attr_init(&attr);
 	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
 	attr_pt= &attr;
-/*
-	libdax_msgs_submit(libdax_messenger, -1, 0x00020158,
-			LIBDAX_MSGS_SEV_DEBUG, LIBDAX_MSGS_PRIO_LOW,
-			"add_worker(): Creating detached thread.", 0, 0);
-*/
-#endif
+
+#endif /* Libburn_create_detached_threadS */
+
+	/* Worker specific locks are to be released early by the worker */
+	if (f == (WorkerFunc) burn_fifo_source_shoveller)
+		burn_async_manage_lock(BURN_ASYNC_LOCK_OBTAIN);
 
 	if (pthread_create(&a->thread, attr_pt, f, a)) {
 		free(a);
@@ -253,7 +294,7 @@ static void reset_progress(struct burn_drive *d, int sessions, int tracks,
 
 int burn_drive_scan(struct burn_drive_info *drives[], unsigned int *n_drives)
 {
-	struct scan_opts o;
+	union w_list_data o;
 	int ret = 0;
 
 	/* ts A61006 : moved up from burn_drive_scan_sync , former Assert */
@@ -266,7 +307,7 @@ int burn_drive_scan(struct burn_drive_info *drives[], unsigned int *n_drives)
 		return -1;
 	}
 
-	/* cant be anything working! */
+	/* cannot be anything working! */
 
 	/* ts A61006 */
 	/* a ssert(!(workers && workers->drive)); */
@@ -292,9 +333,9 @@ drive_is_active:;
 		*drives = NULL;
 		*n_drives = 0;
 
-		o.drives = drives;
-		o.n_drives = n_drives;
-		o.done = 0;
+		o.scan.drives = drives;
+		o.scan.n_drives = n_drives;
+		o.scan.done = 0;
 		add_worker(Burnworker_type_scaN, NULL,
 				 (WorkerFunc) scan_worker_func, &o);
 	} else if (workers->u.scan.done) {
@@ -346,7 +387,7 @@ static void *erase_worker_func(struct w_list *w)
 
 void burn_disc_erase(struct burn_drive *drive, int fast)
 {
-	struct erase_opts o;
+	union w_list_data o;
 
 	/* ts A61006 */
 	/* a ssert(drive); */
@@ -408,8 +449,8 @@ void burn_disc_erase(struct burn_drive *drive, int fast)
 		return;
 	}
 
-	o.drive = drive;
-	o.fast = fast;
+	o.erase.drive = drive;
+	o.erase.fast = fast;
 	add_worker(Burnworker_type_erasE, drive,
 			(WorkerFunc) erase_worker_func, &o);
 }
@@ -447,7 +488,7 @@ static void *format_worker_func(struct w_list *w)
 /* ts A61230 */
 void burn_disc_format(struct burn_drive *drive, off_t size, int flag)
 {
-	struct format_opts o;
+	union w_list_data o;
 	int ok = 0, ret;
 	char msg[40];
 
@@ -535,9 +576,9 @@ void burn_disc_format(struct burn_drive *drive, off_t size, int flag)
 		drive->cancel = 1;
 		return;
 	}
-	o.drive = drive;
-	o.size = size;
-	o.flag = flag;
+	o.format.drive = drive;
+	o.format.size = size;
+	o.format.flag = flag;
 	add_worker(Burnworker_type_formaT, drive,
 			(WorkerFunc) format_worker_func, &o);
 }
@@ -589,7 +630,7 @@ static void *write_disc_worker_func(struct w_list *w)
 
 void burn_disc_write(struct burn_write_opts *opts, struct burn_disc *disc)
 {
-	struct write_opts o;
+	union w_list_data o;
 	char *reasons= NULL;
 	struct burn_drive *d;
 	int mvalid;
@@ -683,9 +724,9 @@ void burn_disc_write(struct burn_write_opts *opts, struct burn_disc *disc)
 
 	d->cancel = 0; /* End of the return = failure area */
 
-	o.drive = d;
-	o.opts = opts;
-	o.disc = disc;
+	o.write.drive = d;
+	o.write.opts = opts;
+	o.write.disc = disc;
 
 	opts->refcount++;
 
@@ -699,7 +740,6 @@ ex:;
 
 static void *fifo_worker_func(struct w_list *w)
 {
-	int old;
 
 #define Libburn_protect_fifo_threaD 1
 
@@ -713,10 +753,6 @@ static void *fifo_worker_func(struct w_list *w)
 	pthread_sigmask(SIG_SETMASK, &sigset, &oldset);
 #endif /* Libburn_protect_fifo_threaD */
 
-	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &old);
-	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &old);
-		/* Note: Only burn_fifo_abort() shall cancel the fifo thread */
-
 	burn_fifo_source_shoveller(w->u.fifo.source, w->u.fifo.flag);
 	remove_worker(pthread_self());
 
@@ -731,7 +767,7 @@ static void *fifo_worker_func(struct w_list *w)
 
 int burn_fifo_start(struct burn_source *source, int flag)
 {
-	struct fifo_opts o;
+	union w_list_data o;
 	struct burn_source_fifo *fs = source->data;
 
 	fs->is_started = -1;
@@ -744,8 +780,8 @@ int burn_fifo_start(struct burn_source *source, int flag)
 		return -1;
 	}
 
-	o.source = source;
-	o.flag = flag;
+	o.fifo.source = source;
+	o.fifo.flag = flag;
 	add_worker(Burnworker_type_fifO, NULL,
 			(WorkerFunc) fifo_worker_func, &o);
 	fs->is_started = 1;
@@ -759,18 +795,19 @@ int burn_fifo_abort(struct burn_source_fifo *fs, int flag)
 	int ret;
 	pthread_t pt;
 
-	if (fs->thread_is_valid <= 0 || fs->thread_handle == NULL)
-		return(2);
+	burn_async_manage_lock(BURN_ASYNC_LOCK_OBTAIN);
 
-#ifdef NIX
-	libdax_msgs_submit(libdax_messenger, -1, 0x00000002,
-			LIBDAX_MSGS_SEV_DEBUG, LIBDAX_MSGS_PRIO_HIGH,
-			"Aborting running burn_source_fifo thread", 0, 0);
-#endif /* NIX */
+	if (fs->thread_is_valid <= 0 || fs->thread_handle == NULL) {
+		burn_async_manage_lock(BURN_ASYNC_LOCK_RELEASE);
+		return 2;
+	}
+	pt = *((pthread_t *) fs->thread_handle);
+
+	burn_async_manage_lock(BURN_ASYNC_LOCK_RELEASE);
+
+	fs->do_abort = 1;
+	ret = pthread_join(pt, NULL);
 
-	pt= *((pthread_t *) fs->thread_handle);
-	remove_worker(pt);
-	ret = pthread_cancel(pt);
 	return (ret == 0);
 }
 
