@@ -2,7 +2,7 @@
 /*
   Class struct of libisoburn.
 
-  Copyright 2007 - 2016 Vreixo Formoso Lopes <metalpain2002@yahoo.es>
+  Copyright 2007 - 2017 Vreixo Formoso Lopes <metalpain2002@yahoo.es>
                     and Thomas Schmitt <scdbackup@gmx.net>
   Provided under GPL version 2 or later.
 */
@@ -710,6 +710,10 @@ struct isoburn_imgen_opts {
     */
     int part_like_isohybrid;
 
+    /* isoburn_igopt_set_iso_mbr_part_type()
+    */
+    int iso_mbr_part_type;
+
     /* See libisoburn.h isoburn_igopt_set_gpt_guid()
     */
     uint8_t gpt_guid[16];
