@@ -35,6 +35,9 @@ struct susp_iterator
     IsoDataSource *src;
     int msgid;
 
+    /* Number of blocks in the ISO 9660 filesystem */
+    uint32_t fs_blocks;
+
     /* block and offset for next continuation area */
     uint32_t ce_block;
     uint32_t ce_off;
@@ -47,7 +50,7 @@ struct susp_iterator
 
 SuspIterator*
 susp_iter_new(IsoDataSource *src, struct ecma119_dir_record *record,
-              uint8_t len_skp, int msgid)
+              uint32_t fs_blocks, uint8_t len_skp, int msgid)
 {
     int pad = (record->len_fi[0] + 1) % 2;
     struct susp_iterator *iter = malloc(sizeof(struct susp_iterator));
@@ -60,6 +63,7 @@ susp_iter_new(IsoDataSource *src, struct ecma119_dir_record *record,
     iter->size = record->len_dr[0] - record->len_fi[0] - 33 - pad;
     iter->src = src;
     iter->msgid = msgid;
+    iter->fs_blocks = fs_blocks;
 
     iter->ce_len = 0;
     iter->buffer = NULL;
@@ -67,12 +71,30 @@ susp_iter_new(IsoDataSource *src, struct ecma119_dir_record *record,
     return iter;
 }
 
-int susp_iter_next(SuspIterator *iter, struct susp_sys_user_entry **sue)
+/* More than 1 MiB in a single file's CE area is suspicious */
+#define ISO_SUSP_MAX_CE_BYTES (1024 * 1024)
+
+
+/* @param flag bit0 = First call on root:
+                      Not yet clear whether this is SUSP at all
+*/
+int susp_iter_next(SuspIterator *iter, struct susp_sys_user_entry **sue,
+                   int flag)
 {
     struct susp_sys_user_entry *entry;
 
     entry = (struct susp_sys_user_entry*)(iter->base + iter->pos);
 
+    if (flag & 1) {
+        /* Yet unclear whether it is SUSP at all */
+        if (iter->size < 7)
+            return 0;
+        if (!SUSP_SIG(entry, 'S', 'P'))
+            return 0;
+        if (entry->len_sue[0] < 7)
+            return 0;
+        /* Looks like SUSP enough to pass the further processing here. */
+    }
     if ( (iter->pos + 4 > iter->size) || (SUSP_SIG(entry, 'S', 'T'))) {
 
         /* 
@@ -81,22 +103,31 @@ int susp_iter_next(SuspIterator *iter, struct susp_sys_user_entry **sue)
          * (IEEE 1281, SUSP. section 4) 
          */
         if (iter->ce_len) {
-            uint32_t block, nblocks;
+            uint32_t block, nblocks, skipped_blocks, skipped_bytes;
 
-            /* A CE has found, there is another continuation area */
-            nblocks = DIV_UP(iter->ce_off + iter->ce_len, BLOCK_SIZE);
+            /* A CE was found, there is another continuation area */
+            skipped_blocks = iter->ce_off / BLOCK_SIZE;
+            skipped_bytes = skipped_blocks * BLOCK_SIZE;
+            nblocks = DIV_UP(iter->ce_off - skipped_bytes + iter->ce_len,
+                             BLOCK_SIZE);
+            if (nblocks <= 0 || iter->ce_len > ISO_SUSP_MAX_CE_BYTES)
+                return ISO_SUSP_WRONG_CE_SIZE;
+            if (((uint64_t) iter->ce_block) + skipped_blocks + nblocks >
+                (uint64_t) iter->fs_blocks)
+                return ISO_SUSP_WRONG_CE_SIZE;
             iter->buffer = realloc(iter->buffer, nblocks * BLOCK_SIZE);
 
-            /* read all blocks needed to cache the full CE */
+            /* Read blocks needed to cache the given CE area range */
             for (block = 0; block < nblocks; ++block) {
                 int ret;
-                ret = iter->src->read_block(iter->src, iter->ce_block + block,
-                                            iter->buffer + block * BLOCK_SIZE);
+                ret = iter->src->read_block(iter->src,
+                                       iter->ce_block + skipped_blocks + block,
+                                       iter->buffer + block * BLOCK_SIZE);
                 if (ret < 0) {
                     return ret;
                 }
             }
-            iter->base = iter->buffer + iter->ce_off;
+            iter->base = iter->buffer + (iter->ce_off - skipped_bytes);
             iter->pos = 0;
             iter->size = iter->ce_len;
             iter->ce_len = 0;
@@ -134,10 +165,10 @@ int susp_iter_next(SuspIterator *iter, struct susp_sys_user_entry **sue)
         }
 
         /* we don't want to return CE entry to the user */
-        return susp_iter_next(iter, sue);
+        return susp_iter_next(iter, sue, 0);
     } else if (SUSP_SIG(entry, 'P', 'D')) {
         /* skip padding */
-        return susp_iter_next(iter, sue);
+        return susp_iter_next(iter, sue, 0);
     }
 
     *sue = entry;
@@ -211,10 +242,19 @@ int read_rr_TF(struct susp_sys_user_entry *tf, struct stat *st)
     
     /* 1. Creation time */
     if (tf->data.TF.flags[0] & (1 << 0)) {
-        
-        /* the creation is the recording time. we ignore this */
-        /* TODO maybe it would be good to manage it in ms discs, where
-         *      the recording time could be different than now!! */     
+        /* Linux accepts ctime by Creation time and by Attributes time.
+         * If both are given, then Attribute time will win.
+         */
+        if (tf->len_sue[0] < 5 + (nts+1) * s) {
+            /* RR TF entry too short. */
+            return ISO_WRONG_RR;
+        }
+        if (s == 7) {
+            time = iso_datetime_read_7(&tf->data.TF.t_stamps[nts*7]);
+        } else {
+            time = iso_datetime_read_17(&tf->data.TF.t_stamps[nts*17]);
+        }
+        st->st_ctime = time;
         ++nts;
     }
     
@@ -365,12 +405,18 @@ int read_rr_SL(struct susp_sys_user_entry *sl, char **dest, int *cont)
         if (*cont == 1) {
             /* new component */
             size_t size = strlen(*dest);
+            int has_slash;
+
             *dest = realloc(*dest, strlen(*dest) + len + 2);
             if (*dest == NULL) {
                 return ISO_OUT_OF_MEM;
             }
             /* it is a new compoenent, add the '/' */
-            if ((*dest)[size-1] != '/') {
+            has_slash = 0;
+            if (size > 0)
+                if ((*dest)[size - 1] == '/')
+                    has_slash = 1;
+            if (!has_slash) {
                 (*dest)[size] = '/';
                 (*dest)[size+1] = '\0';
             }
@@ -412,7 +458,7 @@ int read_rr_PN(struct susp_sys_user_entry *pn, struct stat *st)
 {
     int high_shift= 0;
 
-    if (pn == NULL || pn == NULL) {
+    if (pn == NULL || st == NULL) {
         return ISO_NULL_POINTER;
     }
     if (pn->sig[0] != 'P' || pn->sig[1] != 'N') {
@@ -453,24 +499,27 @@ int read_aaip_AA(struct susp_sys_user_entry *sue,
      if (*is_done) {
 
          /* To coexist with Apple ISO :
-            Gracefully react on eventually trailing Apple AA
+            Gracefully react on possibly trailing Apple AA
          */
          if (sue->version[0] != 1 || sue->len_sue[0] == 7)
              return ISO_SUCCESS;
 
          return ISO_WRONG_RR;
      }
-
-
-     /* Eventually create or grow storage */
      if (*aa_size == 0 || *aa_string == NULL) {
-
-         /* Gracefully react on eventually leading Apple AA
+         /* Gracefully react on possibly leading Apple AA
          */
-         if (sue->version[0] != 1 || sue->len_sue[0] < 9) {
+         if (sue->version[0] != 1 || sue->len_sue[0] < 9)
              return ISO_SUCCESS;
-         }
+     }
 
+     /* A valid AAIP AA entry has 5 header bytes and at least 1 component byte
+      */
+     if (sue->len_sue[0] < 6)
+         return ISO_WRONG_RR;
+
+     /* Possibly create or grow storage */
+     if (*aa_size == 0 || *aa_string == NULL) {
          *aa_size = *aa_len + sue->len_sue[0];
          *aa_string = calloc(*aa_size, 1);
          *aa_len = 0;
@@ -526,7 +575,12 @@ int read_aaip_AL(struct susp_sys_user_entry *sue,
      if (sue->version[0] != 1)
          return ISO_WRONG_RR;
 
-     /* Eventually create or grow storage */
+     /* A valid AL entry has 5 header bytes and at least 1 component byte
+      */
+     if (sue->len_sue[0] < 6)
+         return ISO_WRONG_RR;
+
+     /* Possibly create or grow storage */
      if (*aa_size == 0 || *aa_string == NULL) {
          *aa_size = *aa_len + sue->len_sue[0];
          *aa_string = calloc(*aa_size, 1);
