@@ -1,7 +1,7 @@
 
 /* xorriso - creates, loads, manipulates and burns ISO 9660 filesystem images.
 
-   Copyright 2007-2014 Thomas Schmitt, <scdbackup@gmx.net>
+   Copyright 2007-2016 Thomas Schmitt, <scdbackup@gmx.net>
 
    Provided under GPL version 2 or later.
 
@@ -240,6 +240,7 @@ int Xorriso_node_get_dev(struct XorrisO *xorriso, IsoNode *node,
                bit3= stbuf is to be used without eventual ACL
                bit4= try to obtain a better st_nlink count if hardlinks
                      are enabled
+               bit5= do not look for truncated versions of path component
 */
 int Xorriso_fake_stbuf(struct XorrisO *xorriso, char *path, struct stat *stbuf,
                        IsoNode **node, int flag)
@@ -257,7 +258,8 @@ int Xorriso_fake_stbuf(struct XorrisO *xorriso, char *path, struct stat *stbuf,
    ret= Xorriso_get_volume(xorriso, &volume, 0);
    if(ret<=0)
      return(-1);
-   ret= Xorriso_node_from_path(xorriso, volume, path, node, !(flag&4));
+   ret= Xorriso_node_from_path(xorriso, volume, path, node,
+                               ((flag >> 4) & 2) | !(flag&4));
    if(ret<=0)
      *node= NULL;
  }
@@ -279,7 +281,8 @@ int Xorriso_fake_stbuf(struct XorrisO *xorriso, char *path, struct stat *stbuf,
    stbuf->st_mode|= S_IFLNK;
  else if(LIBISO_ISCHR(*node)) {
    stbuf->st_mode|= S_IFCHR;
-   Xorriso_node_get_dev(xorriso, *node, path, &(stbuf->st_rdev), 0);
+   Xorriso_node_get_dev(xorriso, *node, path, &dev_number, 0);
+   stbuf->st_rdev= dev_number;
  } else if(LIBISO_ISBLK(*node)) {
    stbuf->st_mode|= S_IFBLK;
    /* ts B11124:
@@ -750,7 +753,8 @@ int Xorriso_getfacl(struct XorrisO *xorriso, void *in_node, char *path,
        if(pass) {
          sprintf(*acl_text + strlen(*acl_text), "%s\n", cpt);
        } else if(!(flag & 1)) {
-         Sfile_str(xorriso->result_line, cpt, 0);
+         if(Sfile_str(xorriso->result_line, cpt, 0) <= 0)
+           goto too_long;
          strcat(xorriso->result_line, "\n");
          Xorriso_result(xorriso, 0);
        }
@@ -772,8 +776,10 @@ int Xorriso_getfacl(struct XorrisO *xorriso, void *in_node, char *path,
          } else {
            xorriso->result_line[0]= 0;
            if(what == 0)
-             Sfile_str(xorriso->result_line, "default:", 0);
-           Sfile_str(xorriso->result_line, cpt, 1);
+             if(Sfile_str(xorriso->result_line, "default:", 0) <= 0)
+               goto too_long;
+           if(Sfile_str(xorriso->result_line, cpt, 1) <= 0)
+             goto too_long;
            result_len+= strlen(cpt) + 9;
          }
        } else
@@ -797,6 +803,10 @@ int Xorriso_getfacl(struct XorrisO *xorriso, void *in_node, char *path,
 ex:;
  iso_node_get_acl_text(node, &text, &d_text, 1 << 15);
  return(ret);
+too_long:
+ Xorriso_msgs_submit(xorriso, 0, "Oversized ACL", 0, "FAILURE", 0);
+ ret= 0;
+ goto ex;
 }
 
 
@@ -892,10 +902,12 @@ int Xorriso_append_extattr_comp(struct XorrisO *xorriso,
                                 char *mode, int flag)
 {
  int ret;
+ size_t line_limit;
  char *line, *wpt, *bsl = NULL;
  unsigned char *upt, *uval;
 
  line= xorriso->result_line;
+ line_limit= sizeof(xorriso->result_line);
  uval= (unsigned char *) comp;
 
  if(*mode == 'q') {
@@ -910,6 +922,9 @@ int Xorriso_append_extattr_comp(struct XorrisO *xorriso,
        strcat(line, "\"$(echo -e '");
      wpt= line + strlen(line);
      for(upt= uval; (size_t) (upt - uval) < comp_len; upt++) {
+       if(wpt - line + 5 + 3 + 1 > (ssize_t) line_limit)
+                                                   /* "\\0xyz" + "')\"" + 0 */
+         goto too_much;
        if(*upt <= 037 || *upt >= 0177 || *upt == '\\' || *upt == '\'') {
          if(flag & 1)
            *(wpt++)= '\\';
@@ -929,10 +944,14 @@ int Xorriso_append_extattr_comp(struct XorrisO *xorriso,
    ret= Sfile_bsl_encoder(&bsl, comp, comp_len, 8);
    if(ret <= 0) 
      {ret= -1; goto ex;}
+   if(strlen(line) + strlen(bsl) + 1 > line_limit)
+     goto too_much;
    strcat(line, bsl);
    free(bsl);
    bsl= NULL;
  } else if(*mode == 'r') {
+   if(strlen(line) + strlen(comp) + 1 > line_limit)
+     goto too_much;
    strcat(line, comp);
  }
  ret= 1;
@@ -940,6 +959,11 @@ ex:;
  if(bsl != NULL)
    free(bsl);
  return(ret);
+too_much:;
+ Xorriso_msgs_submit(xorriso, 0, "Oversized BSD-style file attribute",
+                     0, "FAILURE", 0);
+ ret= -1;
+ goto ex;
 }
 
 
@@ -1010,15 +1034,20 @@ int Xorriso_list_extattr(struct XorrisO *xorriso, void *in_node, char *path,
    break;
    if((size_t) (upt - uval) < value_lengths[i]) { 
      strcpy(line, "echo 'OMITTED: Value contains 0-bytes : space \"'\"");
-     Xorriso_append_extattr_comp(xorriso, space_pt, strlen(space_pt), "e", 1);
+     ret= Xorriso_append_extattr_comp(xorriso, space_pt, strlen(space_pt),
+                                      "e", 1);
      if(ret <= 0)
        goto ex;
      strcat(line, "\"'\" , name \"'\"");
-     Xorriso_append_extattr_comp(xorriso, name_pt, strlen(name_pt), "e", 1);
+     ret= Xorriso_append_extattr_comp(xorriso, name_pt, strlen(name_pt),
+                                      "e", 1);
      if(ret <= 0)
        goto ex;
      strcat(line, "\"'\" in file '\"");
-     Xorriso_append_extattr_comp(xorriso, path_pt, strlen(path_pt), "e", 1);
+     ret= Xorriso_append_extattr_comp(xorriso, path_pt, strlen(path_pt),
+                                      "e", 1);
+     if(ret <= 0)
+       goto ex;
      strcat(line, "\" >&2\n");
 
      /* temporarily disable -backslash_codes with result output */
@@ -1369,13 +1398,9 @@ int Xorriso_show_du_subs(struct XorrisO *xorriso, IsoDir *dir_node,
  off_t sub_size, report_size, mem= 0;
  char *path= NULL, *show_path= NULL, *sfe= NULL;
 
- sfe= malloc(5*SfileadrL);
- path= malloc(SfileadrL);
- show_path= malloc(SfileadrL);
- if(path==NULL || show_path==NULL || sfe==NULL) {
-   Xorriso_no_malloc_memory(xorriso, &sfe, 0);
-   {ret= -1; goto ex;}
- }
+ Xorriso_alloc_meM(sfe, char, 5 * SfileadrL);
+ Xorriso_alloc_meM(path, char, SfileadrL);
+ Xorriso_alloc_meM(show_path, char, SfileadrL);
 
  *size= 0;
  ret= iso_dir_get_children(dir_node, &iter);
@@ -1487,12 +1512,9 @@ much_too_long:;
  }
  ret= 1;
 ex:;
- if(sfe!=NULL)
-   free(sfe);
- if(path!=NULL)
-   free(path);
- if(show_path!=NULL)
-   free(show_path);
+ Xorriso_free_meM(sfe);
+ Xorriso_free_meM(path);
+ Xorriso_free_meM(show_path);
  if(iter!=NULL)
    iso_dir_iter_free(iter);
  if(node_array!=NULL)
@@ -1586,10 +1608,91 @@ int Xorriso_node_eff_hidden(struct XorrisO *xorriso, IsoNode *node, int flag)
 }
 
 
+/* @param flag bit0= do not truncate components which contain any of "*?["
+*/
+int Xorriso_truncate_path_comps(struct XorrisO *xorriso, char *path,
+                                char *buffer, char **resultpt, int flag)
+{
+ char *rpt, *basrpt, *wpt, *baswpt, *cpt;
+ int ended, ret, skip;
+
+ *resultpt= path;
+
+ /* Check component lengths */
+ rpt= path;
+ if(*rpt == '/')
+   rpt++;
+ for(ended= 0; !ended;) {
+   basrpt= rpt;
+   rpt= strchr(basrpt, '/');
+   if(rpt == NULL) {
+     rpt= basrpt + strlen(basrpt);
+     ended= 1;
+   }
+   skip= 0;
+   if(flag & 1) {
+     for(cpt= basrpt; cpt < rpt; cpt++) {
+       if(strchr("*?[", *cpt) != NULL) {
+         skip= 1;
+     break;
+       }
+     }
+   }
+   if((!skip) && rpt - basrpt > xorriso->file_name_limit) {
+     ended= 0;
+ break;
+   }
+   if(*rpt == '/')
+     rpt++;
+ }    
+ if(ended)
+   return(1); /* All short enough */
+
+ /* Some truncation is needed */ 
+ buffer[0]= 0;
+ wpt= buffer;
+ if(path[0] == '/')
+   *(wpt++)= '/';
+ rpt= path;
+ if(*rpt == '/')
+   rpt++;
+ for(ended= 0; !ended;) {
+   basrpt= rpt;
+   baswpt= wpt;
+   rpt= strchr(basrpt, '/');
+   if(rpt == NULL) {
+     rpt= basrpt + strlen(basrpt);
+     ended= 1;
+   }
+   skip= 0;
+   for(cpt= basrpt; cpt < rpt; cpt++) {
+     *(wpt++)= *cpt;
+     if((flag & 1) && strchr("*?[", *cpt) != NULL)
+       skip= 1;
+   }
+   *wpt= 0;
+   if((!skip) && rpt - basrpt > xorriso->file_name_limit) {
+      ret= iso_truncate_leaf_name(1, xorriso->file_name_limit, baswpt, 0);
+      Xorriso_process_msg_queues(xorriso, 0);
+      if(ret < 0)
+        return(0);
+      wpt= baswpt + strlen(baswpt);
+   }
+   if(!ended) {
+     *(wpt++)= '/';
+     rpt++;
+   }
+ }
+ *resultpt= buffer;
+ return(1);
+}
+
+
 /* @param flag bit0= long format
                bit1= do not print count of nodes
                bit2= du format
                bit3= print directories as themselves (ls -d)
+               bit5= do not look for truncated versions of path component
 */
 int Xorriso_ls_filev(struct XorrisO *xorriso, char *wd,
                      int filec, char **filev, off_t boss_mem, int flag)
@@ -1598,7 +1701,7 @@ int Xorriso_ls_filev(struct XorrisO *xorriso, char *wd,
  IsoNode *node;
  IsoImage *volume;
  char *path= NULL, *link_target= NULL, *rpt, **dfilev= NULL;
- char *a_text= NULL, *d_text= NULL;
+ char *a_text= NULL, *d_text= NULL, *namept;
  off_t size;
  struct stat stbuf;
 
@@ -1620,7 +1723,8 @@ int Xorriso_ls_filev(struct XorrisO *xorriso, char *wd,
      was_error++;
  continue;
    }
-   ret= Xorriso_node_from_path(xorriso, volume, path, &node, 1);
+   ret= Xorriso_node_from_path(xorriso, volume, path, &node,
+                               1 | ((flag >> 4) & 2));
    if(ret<=0) {
      sprintf(xorriso->info_text, "Not found in ISO image: ");
      Text_shellsafe(path, xorriso->info_text, 1);
@@ -1669,7 +1773,7 @@ int Xorriso_ls_filev(struct XorrisO *xorriso, char *wd,
          Xorriso_result(xorriso,0);
        }
        Xorriso_ls_filev(xorriso, path,
-                        dfilec, dfilev, boss_mem, (flag&1)|2|8);
+                        dfilec, dfilev, boss_mem, (flag & 1) | 2 | 8);
      }
      if(dfilec>0)
        Sfile_destroy_argv(&dfilec, &dfilev, 0);
@@ -1707,12 +1811,13 @@ int Xorriso_ls_filev(struct XorrisO *xorriso, char *wd,
      else
        sprintf(rpt, "%7.f ",(double) (size/1024));
    }
+   Xorriso_truncate_path_comps(xorriso, filev[i], path, &namept, 0);
    if(link_target[0] && (flag&5)==1) {
-     Xorriso_esc_filepath(xorriso, filev[i], xorriso->result_line, 1),
+     Xorriso_esc_filepath(xorriso, namept, xorriso->result_line, 1);
      strcat(xorriso->result_line, " -> ");
      Xorriso_esc_filepath(xorriso, link_target, xorriso->result_line, 1 | 2);
    } else {
-     Xorriso_esc_filepath(xorriso, filev[i], xorriso->result_line, 1);
+     Xorriso_esc_filepath(xorriso, namept, xorriso->result_line, 1);
    }
    strcat(xorriso->result_line, "\n");
    Xorriso_result(xorriso, 0);
@@ -1919,6 +2024,7 @@ ex:;
 /* @param flag bit0= a match count !=1 is a FAILURE event
                bit1= with bit0 tolerate 0 matches if pattern is a constant
                bit2= do not issue debug messages about temporay memory needs
+               bit3= do not add unresolved pattern to filev
 */
 int Xorriso_expand_pattern(struct XorrisO *xorriso,
                            int num_patterns, char **patterns, int extra_filec,
@@ -1987,7 +2093,8 @@ int Xorriso_expand_pattern(struct XorrisO *xorriso,
                                        mem, &dive_count, 1 | abs_adr);
    if(ret<=0)
      goto ex;
-   if(was_count==count && strcmp(patterns[i],"*")!=0 && (flag&3)!=1) {
+   if(was_count==count && strcmp(patterns[i],"*")!=0 && (flag&3)!=1 &&
+      !(flag & 8)) {
      count++;
      Xorriso_eval_nonmatch(xorriso, patterns[i], &nonconst_mismatches, mem, 0);
    }
@@ -1998,9 +2105,9 @@ int Xorriso_expand_pattern(struct XorrisO *xorriso,
  if(ret<=0)
    goto ex;
  count+= extra_filec;
- mem+= extra_filec*sizeof(char *);
+ (*mem)+= extra_filec * sizeof(char *);
  if(count<=0)
-   {ret= 0; goto ex;}
+   {ret= !!(flag & 8); goto ex;}
  ret= Xorriso_alloc_pattern_mem(xorriso, *mem, count, filev, !!(flag & 4));
  if(ret<=0)
    goto ex;
@@ -2017,7 +2124,8 @@ int Xorriso_expand_pattern(struct XorrisO *xorriso,
                                        mem, &dive_count, abs_adr);
    if(ret<=0)
      goto ex;
-   if(was_filec == *filec && strcmp(patterns[i],"*")!=0) {
+   if(was_filec == *filec && strcmp(patterns[i],"*")!=0 && (flag&3) != 1 &&
+      !(flag & 8)) {
      (*filev)[*filec]= strdup(patterns[i]);
      if((*filev)[*filec]==NULL) {
        (*mem)= strlen(patterns[i])+1;
@@ -2030,8 +2138,7 @@ int Xorriso_expand_pattern(struct XorrisO *xorriso,
  ret= 1;
 ex:;
  if(ret<=0) {
-   if(filev!=NULL)
-     Sfile_destroy_argv(&count, filev, 0);
+   Sfile_destroy_argv(&count, filev, 0);
    *filec= 0;
  }
  return(ret);
@@ -2355,9 +2462,10 @@ int Xorriso_file_eval_damage(struct XorrisO *xorriso, IsoNode *node,
 
 
 /* @param flag  bit0= report_sections : section size rather than total size
+                bit1= last_data_file_block : looking for highest data file block
 */
 int Xorriso_report_lba(struct XorrisO *xorriso, char *show_path,
-                       IsoNode *node, int flag)
+                       IsoNode *node, uint32_t *last_block, int flag)
 {
  int ret, *start_lbas= NULL, *end_lbas= NULL, lba_count, i;
  off_t size, *section_sizes= NULL;
@@ -2373,12 +2481,17 @@ int Xorriso_report_lba(struct XorrisO *xorriso, char *show_path,
  for(i= 0; i < lba_count; i++) {
    if(flag & 1)
      size= section_sizes[i];
-   sprintf(xorriso->result_line,
-           "File data lba: %2d , %8d , %8d , %8.f , ",
-           i, start_lbas[i], end_lbas[i] + 1 - start_lbas[i], (double) size);
-   Text_shellsafe(show_path, xorriso->result_line, 1);
-   strcat(xorriso->result_line, "\n");
-   Xorriso_result(xorriso, 0);
+   if(flag & 2) {
+     if(end_lbas[i] > 0 && (uint32_t) end_lbas[i] > *last_block)
+       *last_block= end_lbas[i];
+   } else {
+     sprintf(xorriso->result_line,
+             "File data lba: %2d , %8d , %8d , %8.f , ",
+             i, start_lbas[i], end_lbas[i] + 1 - start_lbas[i], (double) size);
+     Text_shellsafe(show_path, xorriso->result_line, 1);
+     strcat(xorriso->result_line, "\n");
+     Xorriso_result(xorriso, 0);
+   }
  }
  ret= 1;
 ex:;
@@ -2469,7 +2582,9 @@ int Xorriso_search_hardlinks(struct XorrisO *xorriso, IsoNode *node,
 }
 
 
-/* @param flag bit0=do not complain about non existent node */
+/* @param flag bit0=do not complain about non existent node
+               bit1= do not try to find truncated name first
+ */
 int Xorriso_node_from_path(struct XorrisO *xorriso, IsoImage *volume,
                            char *path, IsoNode **node, int flag)
 {
@@ -2485,7 +2600,11 @@ int Xorriso_node_from_path(struct XorrisO *xorriso, IsoImage *volume,
      return(ret);
  }
  *node= NULL;
- ret= iso_tree_path_to_node(volume, path_pt, node);
+ ret= 2;
+ if(!(flag & 2))
+   ret= iso_image_path_to_node(volume, path_pt, node);
+ if(ret == 2)
+   ret= iso_tree_path_to_node(volume, path_pt, node);
  Xorriso_process_msg_queues(xorriso,0);
  if(ret<=0 || (*node)==NULL) {
    if(!(flag&1)) {
@@ -2500,6 +2619,9 @@ int Xorriso_node_from_path(struct XorrisO *xorriso, IsoImage *volume,
 }
 
 
+/* @param flag bit0=do not complain about non existent node
+               bit1= do not try to find truncated name first
+ */
 int Xorriso_dir_from_path(struct XorrisO *xorriso, char *purpose,
                           char *path, IsoDir **dir_node, int flag)
 {
@@ -2511,7 +2633,7 @@ int Xorriso_dir_from_path(struct XorrisO *xorriso, char *purpose,
  if(ret<=0)
    return(ret);
 
- ret= Xorriso_node_from_path(xorriso, volume, path, &node, 0);
+ ret= Xorriso_node_from_path(xorriso, volume, path, &node, flag & 3);
  if(ret<=0)
    goto wdi_is_not_a_dir;
  if(LIBISO_ISDIR(node))
